/*****************************************************************************
 *                                                                           *
 *                    NEC 78K0S/Kx1+ Serial Programmer                       *
 *                            For Windows XP                                 *
 *                                                                           *
 *****************************************************************************
 based on code by ChaN   http://elm-chan.org/

 =============================================================================
 History:
 2013/4/17	V0.0   Created by Bruce Abbott (www.bhabbott.net.nz)
                   Compiled with MinGW32 GCC V3.4.2 under Code::Blocks 12.11

 2013/4/30  V0.1   Replaced MessageBox with custom dialog.

 2013/5/07  V0.2   Hex display scrollbar adjusted to match ROM size.
                   Settings saved to registry on exit.

 2013/5/12  V0.3   Erase blocks, Write blocks, scale on progress bar.

 2013/5/17  V0.4   Byte Map (shows which bytes in a block are blank).

 2013/5/23  V0.5   Hex display color enhancments.

 2013/5/28  V0.6   Progress bar color enhancements.

                   Write Blocks now skips blank blocks. Also skips leading and
                   trailing blank bytes in each block. Now you can write bytes
                   into a block which is not blank!

                   Added "\\\\.\\" to beginning of com port name, to enable
                   opening COM10~99.

 2013/6/16 V0.7    Power on/off and application icons.

 =============================================================================
 */
//#define simulate       // don't write or erase chip - for testing only!

#define WINVER 0x0501   // Windows XP or later
#define _WIN32_IE 0x0501 // IE 5.01+ (required for InitCommonControlsEx)

// Use Unicode 16 bit 'wide' characters and strings (UTF-16)
#define UNICODE
#define _UNICODE

#define WIN32_LEAN_AND_MEAN // exclude less-common APIs
#include <windows.h>
#include <stdio.h>		// standard I/O functions (file, comm port)
#include <commctrl.h>	// common controls (progress bar)
#include <commdlg.h>    // GetOpenFilename dialog

// menu item identifiers
#define IDM_FILE_OPEN 1	// open hex file

// window identifiers
#define	IDW_MAIN	100	// main window
#define IDW_REQW    200 // requester window
#define IDW_HEXD	300	// hex display window
#define	IDW_PROG	400	// programming window

// main buttons
#define	IDB_COM		101	// com port dropdown box
#define	IDB_MEM		102	// memory size selection group box
#define	IDB_1K		103	// select 1k radio button
#define	IDB_2K		104	// select 2k radio button
#define	IDB_4K		105	// select 4k radio button
#define	IDB_TEST	106	// test comms button
#define IDB_BLANK   107 // blank check button
#define	IDB_ERASE	108	// erase flash button
#define	IDB_WRITE	109	// write flash button
#define IDB_QUIT    110 // quit button

// requester buttons
#define IDQ_OK      201 // OK button
#define IDQ_CANCEL  202 // Cancel button
#define IDQ_RETRY   203 // Retry button

// write flash buttons
#define	IDB_WCHP	401 // write chip button
#define IDB_WBLK    402 // write blocks button
#define	IDB_EB4W	403	// erase before write checkbox
#define	IDB_EXIT	404	// close programming window

// erase flash buttons
#define IDB_EBLK    420 // erase blocks button
#define IDB_ECHP    421 // erase chip button

#define IDB_BLKS    499 // blocks groupbox
// block to write/erase checkboxes
#define IDB_BLK0    500 // block 0
//                  ... // blocks 1-14
#define IDB_BLK15   515 // block 15

// messagebox icons
#define MB_ICON_SW_ON  MB_USERICON
#define MB_ICON_SW_OFF (MB_USERICON+16)

/*------------------------ function prototypes -----------------------------*/
void Truncate (wchar_t *s, int maxlen);
DWORD get_valh (char **lp, int digits, BYTE *sum);
long LoadFile (wchar_t* file);
long input_hexfile (FILE *fp, BYTE *buffer, DWORD buffsize, DWORD *datasize);
BOOL send_com (const BYTE* buff, DWORD bytes);
BYTE recv_com ();
int  EnterProgMode ();
void ExitProgMode ();
long LoadFile (wchar_t* file);
int  chiperase ();
int  MapBytes (int block);
int  blankcheck (int block);
int  blockerase (int block);
int  WriteBlock (int block, int first, int last);
void EraseBlocks (HWND hwnd);
void EraseChip (HWND hwnd);
void ShowProgBar (HDC hdc);
void ShowBarSegment (HDC hdc, int block);
void OpenFileDialog (HWND hwnd);
int  TestComms (HWND hwnd);
int  OpenComPort (HWND hwnd);
void DoMap (HWND hwnd, int block);
void WriteBlocks(HWND hwnd);
void WriteChip (HWND hwnd);
int  DoBlank (HWND hwnd);
int  DoErase (HWND hwnd);
int  DoWrite (HWND hwnd);
int  Request (HWND hwnd, LPCWSTR text, LPCWSTR title, INT x, INT y, INT type);
int  SaveSettings ();
int  LoadSettings ();

// wide string compare
static inline int strcmpw (const wchar_t *str1, const wchar_t *str2)
{
    while (*str1 && (*str1 == *str2)) { str1++; str2++; }
    return *str1 - *str2;
}

/*------------------------- global variables -------------------------------*/
HINSTANCE g_hinst;		            // global copy of hInstance

DWORD CodeSize = 0;		 		    // size of code extracted from hex file
DWORD flashsize = 4096;			    // size of target MPU's flash memory
BYTE CodeBuff[4096];	 		    // binary code to be written to flash

INT blockstates[16];                // erasure status of FlashROM blocks
INT blocks[16];                     // blocks to be written or erased
INT blocksdone[16];                 // blocks shown as done in progress bar

BYTE bytemap[256];                  // map of erased (0xFF) bytes in a block
int mapblock = 0;                   // block to be mapped

int eraseflag = 0;                  // 1 if 'erase before write' selected

int window_x = CW_USEDEFAULT;       // main window left coordinate
int window_y = CW_USEDEFAULT;       // main window top coordinate

// window handles
HWND mainwindow;		            // main window
HWND hexwindow;	            		// hexadecimal display
HWND writewindow;		            // write to flash window
HWND erasewindow;		            // erase blocks window
HWND filenamebox;                   // file name display area
HWND reqwindow;                     // Requester window
HWND blankwindow;                   // Blank Check window
HWND mapwindow;                     // block byte map window

HDC memdc;                          // button bitmap device context
HBITMAP redbmp, grnbmp;             // button bitmaps

// control handles
HWND combox;		            	// com port dropdown box
HWND ProgBar;		            	// programming progress bar

// fonts
HFONT   sanfont = NULL; // san serif font for normal text
HFONT   hexfont = NULL; // fixed width font for hex display

wchar_t comportname[10] = L"COMxx"; // com port name

DCB dcb = {sizeof(DCB),			    // com port device control block
			115200, TRUE, FALSE, FALSE, FALSE,
			DTR_CONTROL_ENABLE, FALSE,
			TRUE, FALSE, FALSE, FALSE, FALSE,
			RTS_CONTROL_ENABLE, FALSE, 0, 0,
			10, 10, 8, EVENPARITY, ONESTOPBIT,
			'\x11', '\x13', '\xFF', '\xFF', 0 };

COMMTIMEOUTS commtimeouts1 = {0,1,200,1,200};	// com port timeout values

HANDLE hComm = INVALID_HANDLE_VALUE; 			// com port handle

wchar_t filename[MAX_PATH] = {L"no file"};      // path/name of hex file

static const wchar_t *blocknames[] = {L"0", L"1", L"2", L"3", L"4", L"5", L"6", L"7",
                                      L"8", L"9", L"10", L"11", L"12", L"13", L"14", L"15"};
static const wchar_t *kbytenames[] = {L"0", L"1k", L"2k", L"3k", L"4k"};

/*----------------------------------------------------------------------------
                      Truncate Long Path/Filename
------------------------------------------------------------------------------
 input: pointer to string to truncate, maximum length of truncated string
output: Keep first 3 chars, insert "...", fill rest with right end of string

eg: input: "C:\Documents and Settings\usr\My Documents\src\ServoReverse\a.hex"
   output: "C:\...y Documents\src\ServoReverse\a.hex" (truncated to 40 chars)

Notes:
1. String is modified in situ! Truncated string may not be a valid filename!
2. maxlen must be >= 6! (probably should be at least 20 for readability)
3. String will not be modified if already equal to or shorter than maxlen.
*/
void Truncate (wchar_t *s, int maxlen)
{
    int i;
    wchar_t *p = s;
    int trim = lstrlenW(p) - maxlen;
    if (trim > 0) {
        p[3] = p[4] = p[5] = L'.';
        for (i = 6; i <= maxlen; i++) {
	        p[i] = p[i+trim];
        }
    }
}

//-----------------------------------------------------------------------------
//              Pick a hexdecimal value from hex record
//-----------------------------------------------------------------------------
DWORD get_valh (char **lp,	// pointer to line read pointer
			   int digits, 	// number of digits to get (2,4,6,8)
			   BYTE *sum)	// byte check sum
{
	DWORD val = 0;
	BYTE n;

	do {
		n = *(*lp)++;
		if ((n -= '0') >= 10) {
			if ((n -= 7) < 10) return(0xFFFFFFFF);
			if (n > 0xF) return(0xFFFFFFFF);
		}
		val = (val << 4) + n;
		if (digits & 1) *sum += (BYTE)val;
	} while (--digits);
	return val;
}

//=============================================================================
//       Load Intel/Motorola hex file and convert to binary
//=============================================================================
// return code: 0 = bad file!

long input_hexfile (FILE *fp,			// input stream
				   BYTE *buffer,		// data input buffer
				   DWORD buffsize,		// size of data buffer
				   DWORD *datasize)		// effective data size in the input buffer
{
	char line[600];			// line input buffer
	char *lp;				// line read pointer
	long lnum = 0;			// input line number
	WORD seg = 0, hadr = 0;	// address expansion values for intel hex
	DWORD addr, bytes, n;
	BYTE sum;

	while (fgets(line, sizeof(line), fp) != NULL) {
		lnum++;
		lp = &line[1]; sum = 0;

		if (line[0] == ':') {	// Intel Hex format
			if ((bytes = get_valh(&lp, 2, &sum)) > 0xFF) return lnum;	// number of bytes
			if ((addr = get_valh(&lp, 4, &sum)) > 0xFFFF) return lnum;	// offset

			switch (get_valh(&lp, 2, &sum)) {	// block type?
			case 0x00:	// data block
				addr += (seg << 4) + (hadr << 16);
				while (bytes--) {
					if ((n = get_valh(&lp, 2, &sum)) > 0xFF) return lnum;
					if (addr >= buffsize) continue;	// clip by buffer size
					buffer[addr++] = (BYTE)n;		// store the data
					if (addr > *datasize)
						*datasize = addr;           // update data size
				}
				break;

			case 0x01:	// end block
				if (bytes) return lnum;
				break;

			case 0x02:	// segment block
				if (bytes != 2) return lnum;
				if ((seg = (WORD)get_valh(&lp, 4, &sum)) == 0xFFFF) return lnum;
				break;

			case 0x04:	//high address block
				if (bytes != 2) return lnum;
				if ((hadr = (WORD)get_valh (&lp, 4, &sum)) == 0xFFFF) return lnum;
				break;

			default:    // invalid block
				return lnum;
			}
			if (get_valh(&lp, 2, &sum) > 0xFF) return lnum; // get checksum
			if (sum) return lnum;			     			// test checksum
		}

		if (line[0] == 'S') {	                        // Motorola S format
			if ((*lp >= '1')&&(*lp <= '3')) {

				switch (*lp++) {	                    // record type? (S1/S2/S3)
				case '1':
					if ((bytes = get_valh(&lp, 2, &sum) - 3) > 0xFF) return lnum;
					if ((addr = get_valh(&lp, 4, &sum)) == 0xFFFFFFFF) return lnum;
					break;
				case '2':
					if ((bytes = get_valh(&lp, 2, &sum) - 4) > 0xFF) return lnum;
					if ((addr = get_valh(&lp, 6, &sum)) == 0xFFFFFFFF) return lnum;
					break;
				default:
					if ((bytes = get_valh(&lp, 2, &sum) - 5) > 0xFF) return lnum;
					if ((addr = get_valh(&lp, 8, &sum)) == 0xFFFFFFFF) return lnum;
				}
				while (bytes--) {
					if ((n = get_valh(&lp, 2, &sum)) > 0xFF) return lnum;
					if (addr >= buffsize) continue;	    // clip by buffer size
					buffer[addr++] = (BYTE)n;		    // store the data
					if (addr > *datasize)			    // update data size information
						*datasize = addr;
				}
				if (get_valh(&lp, 2, &sum) > 0xFF) return lnum;	// get checksum
				if (sum != 0xFF) return lnum;   	  	        // test checksum
			}
		}

	}
	return (feof(fp) ? 0 : -1);
}

/*-----------------------------------------------------------------------------
                          Serial Port Transmit
-------------------------------------------------------------------------------
 Transmits 1 to 4 bytes.
   Command strings are 4 bytes long.

 Receives and consumes the echo.
   Only 1 wire is used for both transmit and receive, so all transmitted bytes
   should echo back (if not then either there was collision or the interface is
   not working).
*/
BOOL send_com (const BYTE* buff, DWORD bytes)
{
	DWORD tc;       // number of bytes actually sent
	BYTE rets[4];   // receive buffer for echo

	WriteFile(hComm, buff, bytes, &tc, NULL);	 // transmit bytes
	ReadFile(hComm, rets, bytes, &tc, NULL);	 // receive echo
	if(tc != bytes) return FALSE;				 // each byte should echo back
	if(memcmp(buff, rets, bytes)) return FALSE;  // echo should equal original
	return TRUE;
}

/*-----------------------------------------------------------------------------
                          Serial Port Receive
-------------------------------------------------------------------------------
   Receives one byte from serial port
*/
BYTE recv_com ()
{
	DWORD bytes;    // number bytes actually read
	BYTE rv = 0;    // temporary buffer to store received byte

	ReadFile(hComm, &rv, 1, &bytes, NULL);  // get 1 byte from serial port
	return rv;      // return the byte that was read
}

/*-------------------------------------------------------------------------
                  Enter Flash Memory Programming Mode
---------------------------------------------------------------------------
Attempts to put the 78K0S/Kx1+ chip into programming mode

Returns TRUE if successful
Returns FALSE if no response from chip or serial port not open

Procedure for entering programming mode:-
 1. Hold RESET low, CLOCK and DATA high.
 2. Pulse CLOCK low once, then pulse DATA low 5 times (by sending 0x55 on TXD)
 3. Release RESET and enable 8MHz clock
 4. Wait for response on RXD
 5. If byte received is 0x01 then we are now in programming mode!
*/
int EnterProgMode ()
{
	BYTE buff[4];                        // buffer to hold command string

    if(!hComm) return FALSE;             // abort if no serial port
	EscapeCommFunction(hComm, SETRTS);   // force RESET Low
	Sleep(20);
	PurgeComm(hComm, PURGE_RXCLEAR);
	EscapeCommFunction(hComm, SETDTR);	 // CLOCK = Low
	EscapeCommFunction(hComm, CLRDTR);	 // CLOCK = High
	buff[0] = 0x55;                      // 0x55 = L+HLHLHLHL+H
	send_com(buff, 1);					 // apply 5 negative pulses to DATA
	Sleep(20);
	EscapeCommFunction(hComm, CLRRTS);	 // release RESET
	Sleep(20);
	EscapeCommFunction(hComm, SETDTR);   // start 8MHz clock
	Sleep(100);

	// send 4 nulls, then wait for reply
	buff[0] = buff[1] = buff[2] = buff[3] = 0x00;
	if (!send_com(buff, 4)) return FALSE;
    if (recv_com() != 0x01) return FALSE; // 0x01 = "in programming mode"

	return TRUE;                          // success!
}

//-----------------------------------------------------------------------
//                     Exit Pogramming Mode
//-----------------------------------------------------------------------
void ExitProgMode ()
{
    if (hComm) {
	EscapeCommFunction(hComm, CLRDTR);   // force CLOCK High
	EscapeCommFunction(hComm, SETRTS);   // force RESET Low
    Sleep(20);
    }
}

/*-----------------------------------------------------------------------
            	            Blank Check
-----------------------------------------------------------------------*/
/*
   Check block for erasure

   return code: 0 = error
                1 = blank
                2 = not blank
*/
int blankcheck(int block)
{
	BYTE c_era[4], res;
	c_era[0] = 0x32;				    // command = "blank check"
	c_era[1] = block;				    // block to check
	c_era[2] = 0x00; c_era[3] = 0xFF;	// entire block
	if (!send_com(c_era, 4)) return 0;	// send 4 byte command string
	if (recv_com() != 0x06) return 0;	// if not ACK then abort
	res = recv_com();				    // receive blank check status
	if (res == 0x06) return 1;          // blank
	return 2;                           // not blank
}

/*-----------------------------------------------------------------------
                             Map Bytes
-----------------------------------------------------------------------*/
/*
  Requests a 'blank check' for each BYTE in a block.
  Results stored in bytemap[]. 0xff = blank, 0x00 = not blank.
  Return code: 0 = OK, -1 = error.
*/
int MapBytes (int block)
{
	BYTE c_era[4];   // 4 byte command string sent to target
	BYTE res;	     // 1 byte response received from target
	int offset = 0;
    memset(bytemap, 0, sizeof(bytemap));
    if (EnterProgMode()) {
        for (; offset < 256; offset++) {
            c_era[0] = 0x32;				        // command = "blank check"
            c_era[1] = block;				        // block to check
            c_era[2] = offset; c_era[3] = offset;	// offset to test
            if (!send_com(c_era, 4)) break; 	    // send 4 byte command string
            if (recv_com() != 0x06) break; 	        // if not ACK then abort
            res = recv_com();				        // receive blank check status
            if (res == 0x06) {
                bytemap[offset] = 0xFF; 	 // Byte = 0xFF (all bits unprogrammed)
            } else {
                bytemap[offset] = 0x00;      // Not blank, and that's all we know!
            }
        }
        ExitProgMode();
    }
	if (offset !=256) return -1;     // error all bytes not mapped
	return 0;
}

/*-----------------------------------------------------------------------
                              Block Erase
-------------------------------------------------------------------------
 Attempts to erase one block.
 Returns number of attempts required, or 0 if erase failed.
*/
int blockerase (int block)
{
#ifdef simulate
    return 1;
#endif
	BYTE c_era[4];                      // 4 byte tx buffer
	BYTE res;                           // 1 byte receive buffer
    int n;
    // set the block number
	c_era[1] = block;
	// offset = 0x00, last address = 0xFF (ie. entire 256 byte block)
	c_era[2] = 0; c_era[3] = 0xFF;
	for (n = 1; n<=25; n++) {           // up to 25 erasure attempts
        // erase block
		c_era[0] = 0x22;				// command = "block erase"
		if (!send_com(c_era, 4)) break;	// send 4 byte command string
		if (recv_com() != 0x06) break;	// if not ACK then abort
		res = recv_com();				// wait until erase complete
		if (res != 0x06) break;			// if not ACK then abort
        // verify block erasure
		c_era[0] = 0x32;                // command = "block erase verify"
		if (!send_com(c_era, 4)) break;	// send 4 byte command string
		if (recv_com() != 0x06) break;	// if not ACK then abort
		res = recv_com();				// response received when verify complete
		if (res == 0x1A) continue;	    // if "erase verify error" then try again
		if (res == 0x06) return n;      // if ACK then chip erased successfully
	}
	return 0;                           // erase failed!
}

/*-----------------------------------------------------------------------
                              Chip Erase
-------------------------------------------------------------------------
 Attempts to erase the entire chip.
 Returns number of attempts required, or zero if erase failed.
*/
int chiperase ()
{
#ifdef simulate
    return 1;
#endif
	BYTE c_era[4], res;
	BYTE max_page = (BYTE)((flashsize >> 8)-1);
    int n;
	// offset = 0x00, last address = 0xFF (ie. entire 256 byte block)
	c_era[2] = 0; c_era[3] = 0xFF;
	for (n = 1; n<=10; n++) {           // up to 10 erasure attempts
        // erase chip
		c_era[0] = 0x20;				// command = "chip erase"
		c_era[1] = max_page;			// erase all blocks
		if (!send_com(c_era, 4)) break;	// send 4 byte command string
		if (recv_com() != 0x06) break;	// if not ACK then abort
		res = recv_com();				// wait until erase complete
		if (res != 0x06) break;			// if not ACK then abort
        // verify chip erasure
		c_era[0] = 0x30;				// command = "chip erase verify"
		if (!send_com(c_era, 4)) break;	// send 4 byte command string
		if (recv_com() != 0x06) break;	// if not ACK then abort
		res = recv_com();				// response received when verify complete
		if (res == 0x1A) continue;		// if "erase verify error" then try again
		if (res != 0x06) break;			// if not ACK then abort
        // verify erasure of block 0x80 (security byte)
		c_era[0] = 0x32;                // command = "block erase verify"
		c_era[1] = 0x80;				// block = 0x80
		if (!send_com(c_era, 4)) break;	// send 4 byte command string
		if (recv_com() != 0x06) break;	// if not ACK then abort
		res = recv_com();				// response received when verify complete
		if (res == 0x1A) continue;	    // if "erase verify error" then try again
		if (res == 0x06) return n;      // if ACK then chip erased successfully
	}
	return 0;       // erase failed!
}

/*-------------------------------------------------------------------------------
                                Open Com Port
 --------------------------------------------------------------------------------
 return code: 1 = com port opened OK,
              0 = com port not opened
*/

int OpenComPort (HWND hwnd)
{
    wchar_t comfilename[20] = L"\\\\.\\";
    lstrcatW(comfilename, comportname);
    hComm = CreateFileW(comfilename, GENERIC_READ|GENERIC_WRITE, 0, NULL,
      	                      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hComm == INVALID_HANDLE_VALUE) {
        Request(hwnd, L"Could not open Com port!", L"78KOS/Kx1 Prog - Error!",
                                            210, 120, MB_ICONERROR | MB_OK);
        return 0;
    } else {
        SetCommState(hComm, &dcb);
        SetCommTimeouts(hComm, &commtimeouts1);
        return 1;
    }
}



#define ECX 80
#define ECY 15
/*-----------------------------------------------------------------------
                   "Erase Chip" Button Pressed
-------------------------------------------------------------------------
*/
void EraseChip (HWND hwnd)
{
    int rc;
    HCURSOR busy_pointer = LoadCursorW(NULL, IDC_WAIT); // get busy pointer (hourglass)

 	EnableWindow(GetDlgItem(hwnd, IDB_EXIT), FALSE);
	EnableWindow(GetDlgItem(hwnd, IDB_EB4W), FALSE);    // disable buttons while erasing
	EnableWindow(GetDlgItem(hwnd, IDB_EBLK), FALSE);

    SetCursor(busy_pointer);
    do {
        if (!EnterProgMode()) {                 // attempt to enter programming mode
            rc = Request(hwnd, L"Target not responding!", L"Erase Chip - Error!",
                                        ECX, ECY, MB_ICONWARNING|MB_RETRYCANCEL);
        } else {
            rc = (chiperase());				    // attempt to erase chip
            SetCursor(NULL);
            if (rc == 0) {
                rc = Request(hwnd, L"Erase failed!", L"Erase Chip - Error!",
                                            ECX, ECY, MB_ICONWARNING|MB_OK);
            } else {
                Sleep(200);
                rc = Request(hwnd, L"Chip is now erased", L"Erase Chip - Passed",
                                                               ECX, ECY, MB_OK);
            }
        }
        ExitProgMode();
    } while (rc == IDRETRY);
    SetCursor(NULL);
 	EnableWindow(GetDlgItem(hwnd, IDB_EXIT), TRUE);
	EnableWindow(GetDlgItem(hwnd, IDB_EB4W), TRUE);    // enable buttons after erase
	EnableWindow(GetDlgItem(hwnd, IDB_EBLK), TRUE);
}

#define EBRX 125
#define EBRY 15
/*-----------------------------------------------------------------------
                   "Erase Blocks" Button Pressed
-------------------------------------------------------------------------
*/
void EraseBlocks (HWND hwnd)
{
    int rc;         // return code
    int block;      // current block

    HCURSOR busy_pointer = LoadCursorW(NULL, IDC_WAIT); // get busy pointer (hourglass)

 	EnableWindow(GetDlgItem(hwnd, IDB_EXIT), FALSE);
	EnableWindow(GetDlgItem(hwnd, IDB_EB4W), FALSE);    // disable buttons while erasing
	EnableWindow(GetDlgItem(hwnd, IDB_ECHP), FALSE);

    rc = 0;
    for (block = 0; block < (flashsize/256); block++) {
        if (IsDlgButtonChecked(hwnd, IDB_BLK0+block) == BST_CHECKED) {
            blocks[block] = 1;
            rc++;               // count number of blocks to write
        } else {
            blocks[block] = 0;
        }
    }
    if (rc == 0) {
       Request(hwnd,L"No blocks selected", L"Erase Blocks",
               EBRX, EBRY, MB_ICONINFORMATION | MB_OK);
    } else {
        SetCursor(busy_pointer);
        do {
            if (!EnterProgMode()) {                 // attempt to enter programming mode
                rc = Request(hwnd, L"Target not responding!", L"Erase Blocks - Error!",
                                            EBRX, EBRY, MB_ICONWARNING|MB_RETRYCANCEL);
            } else {
                for (block = 0; block < (flashsize/256); block++) {
                    if (blocks[block] == 1) {           // if this block is to be erased
                        rc = (blockerase(block));	    // attempt to erase block
                        if (rc == 0) {
                            Request(hwnd, L"Erase failed!", L"Erase Blocks - Error!",
                                                   EBRX, EBRY, MB_ICONWARNING|MB_OK);
                            break;
                        }
                    }
                }
                if (rc != 0) {
                    Sleep(200);
                    rc = Request(hwnd, L"Selected blocks are now erased", L"Erase Blocks",
                                                                               EBRX, EBRY, MB_OK);
                }
            }
            ExitProgMode();
        } while (rc == IDRETRY);
        SetCursor(NULL);
    }
 	EnableWindow(GetDlgItem(hwnd, IDB_EXIT), TRUE);
	EnableWindow(GetDlgItem(hwnd, IDB_EB4W), TRUE);    // enable buttons after erase
	EnableWindow(GetDlgItem(hwnd, IDB_ECHP), TRUE);
}


/*-----------------------------------------------------------------------
          Write One Block (256 Bytes) to Flash Memory
-------------------------------------------------------------------------
 input:
     block = block to be written (FlashROM address / 256)
     first = offset in block of first byte to be written (0~255)
     last  = offset in block of last byte to be written  (0~255)

 returns TRUE if successful, FALSE if failed
*/
int WriteBlock (int block, int first, int last)
{
#ifdef simulate
    Sleep(250);
    return TRUE;
#endif
	BYTE c_prg[4];                           // 4 byte command string
    int offset;
	c_prg[2] = first; c_prg[3] = last;       // set first/last byte offsets
	c_prg[0] = 0x40;                         // command = "write block"
	c_prg[1] = block;                        // set block number
    // write
	if (!send_com(c_prg, 4)) return FALSE;   // send command
    if (recv_com() != 0x06) return FALSE;    // ACK = command received
    for (offset = first; offset <= last; offset++) {
        send_com(&CodeBuff[(block*256)+offset], 1); // send one data byte
        if (recv_com() != 0x06) return FALSE;       // ACK = byte received OK
       }
    if (recv_com() != 0x06) return FALSE;    // ACK = block programmed
    // verify
    c_prg[0] = 0x19;                         // command = "internal verify"
    if (!send_com(c_prg, 4)) return FALSE;   // send command
    if (recv_com() != 0x06) return FALSE;    // 1st ACK = Verify started
    if (recv_com() != 0x06) return FALSE;    // 2nd ACK = Write Verify OK
    return TRUE;
}

#define SWX 128
#define SWY 68
//-----------------------------------------------------------------------------
//                     'Write Chip' button pressed
//-----------------------------------------------------------------------------
void WriteChip (HWND hwnd)
{
    HDC hdc;
    HCURSOR busy_pointer;
    int rc;    // return code
	int	firstblock = 0, lastblock = (flashsize >> 8)-1; // all FlashROM code blocks
    int block; // current block

    busy_pointer = LoadCursorW(NULL, IDC_WAIT);

    for (block = 0; block < 16; block++) {
        blocksdone[block] = 0;
    }
    EnableWindow(GetDlgItem(hwnd, IDB_EXIT), FALSE);
    EnableWindow(GetDlgItem(hwnd, IDB_EB4W), FALSE);    // disable buttons while writing
    EnableWindow(GetDlgItem(hwnd, IDB_WBLK), FALSE);
	EnableWindow(GetDlgItem(hwnd, IDB_BLKS), FALSE);
    do {
        hdc = GetDC(hwnd);
        ShowProgBar(hdc);                       // refresh progress bar
        ReleaseDC(hwnd, hdc);
        SetCursor(busy_pointer);                // show busy pointer (hourglass)
        if (!EnterProgMode()) {                 // attempt to enter programming mode
            SetCursor(NULL);
            rc = Request(hwnd, L"Target not responding!", L"Write Chip - Error!",
                                        SWX, SWY, MB_ICONWARNING|MB_RETRYCANCEL);
        } else {
            if (eraseflag) {                    // "erase before write" option checked?
                rc = (chiperase());				// yes, attempt to erase chip
            } else {
                rc = IDIGNORE;                 // no, ignore erase state
            }
            if (rc == 0) {
                SetCursor(NULL);
                rc = Request(hwnd, L"Erase failed!", L"Write Chip - Error!",
                                            SWX, SWY, MB_ICONWARNING|MB_OK);
            } else {
                rc = IDIGNORE;                        // return code = don't retry
                for (block = firstblock; block <= lastblock; block++) {
                    if (WriteBlock(block, 0, 255)) {  // write block to FlashROM
                        blocksdone[block] = 2;
                        hdc = GetDC(hwnd);
                        ShowBarSegment(hdc, block);   // show block written in progress bar
                        ReleaseDC(hwnd, hdc);
                    } else {
                        SetCursor(NULL);
                        rc = Request(hwnd, L"Write failed!", L"Write Chip - Error!",
                                                    SWX, SWY, MB_ICONWARNING|MB_OK);
                        blocksdone[block] = 3;       // block write failed
                        hdc = GetDC(hwnd);
                        ShowBarSegment(hdc, block);  // show failed block in progress bar
                        ReleaseDC(hwnd, hdc);
                        break;
                    }
                }
                SetCursor(NULL);                // show normal pointer
            }
        }
       ExitProgMode();
    } while (rc == IDRETRY);

    EnableWindow(GetDlgItem(hwnd, IDB_BLKS), TRUE);
    EnableWindow(GetDlgItem(hwnd, IDB_WBLK), TRUE);
    EnableWindow(GetDlgItem(hwnd, IDB_EB4W), TRUE);     // enable buttons after writing
    EnableWindow(GetDlgItem(hwnd, IDB_EXIT), TRUE);
}

//-----------------------------------------------------------------------------
//                     'Write Blocks' button pressed
//-----------------------------------------------------------------------------
void WriteBlocks (HWND hwnd)
{
    HDC hdc;
    HCURSOR busy_pointer;
    int rc;             // return code
    int block;          // current block
    int first, last;    // start and end offset address in current block

    busy_pointer = LoadCursorW(NULL, IDC_WAIT); // get busy pointer (hourglass)

 	EnableWindow(GetDlgItem(hwnd, IDB_EXIT), FALSE);
	EnableWindow(GetDlgItem(hwnd, IDB_EB4W), FALSE);    // disable buttons while writing
	EnableWindow(GetDlgItem(hwnd, IDB_WCHP), FALSE);
    rc = 0;
    for (block = 0; block < 16; block++) {
        blocksdone[block] = 0;
        if (IsDlgButtonChecked(hwnd, IDB_BLK0+block) == BST_CHECKED) {
            blocks[block] = 1; // block to be written
            rc++;
        } else {
            blocks[block] = 0; // block to be skipped
        }
    }
    if (rc == 0) {
        Request(hwnd,L"No blocks selected", L"Write Blocks", SWX, SWY,
                MB_ICONINFORMATION | MB_OK);
    } else {
        SetCursor(busy_pointer);                    // show busy mouse pointer
        do {
            hdc = GetDC(hwnd);
            ShowProgBar(hdc);                       // refresh progress bar
            ReleaseDC(hwnd, hdc);
            if (!EnterProgMode()) {                 // attempt to enter programming mode
                SetCursor(NULL);
                rc = Request(hwnd, L"Target not responding!", L"Write Blocks - Error!",
                             SWX, SWY, MB_ICONWARNING|MB_RETRYCANCEL);
            } else {
                rc = IDIGNORE;                         // return code = don't retry
                for (block = 0; block < (flashsize/256); block++) {
                    if (blocks[block] == 1) {          // if this block is to be written
                        if (eraseflag) {               // "erase before write" option checked?
                            rc = (blockerase(block));  // yes, attempt to erase block
                        } else {
                            rc = IDIGNORE;             // no, ignore erase state
                        }
                        if (rc == 0) {
                            SetCursor(NULL);
                            rc = Request(hwnd, L"Erase failed!", L"Write Blocks - Error!",
                                         SWX, SWY, MB_ICONWARNING|MB_OK);
                            break;
                        } else {
                            for (first = 0; first < 256; first++) {
                                if (CodeBuff[(block*256)+first] != 0xFF) break;  // skip blank bytes at begining
                            }
                            if (first < 256) {         // if at least 1 byte to write
                                for (last = 255; last >= first; last--) {
                                    if (CodeBuff[(block*256)+last] != 0xFF) break;  // skip blank bytes at end
                                }
                                if (WriteBlock(block, first, last)) {  // write the block
                                } else {
                                    rc = Request(hwnd, L"Write failed!", L"Write Blocks - Error!",
                                                        SWX, SWY, MB_ICONWARNING|MB_OK);
                                    blocksdone[block] = 3;       // block write failed
                                    hdc = GetDC(hwnd);
                                    ShowBarSegment(hdc, block);  // show failed block in prgress bar
                                    ReleaseDC(hwnd, hdc);
                                    break;
                                        }
                            } else {
                                Sleep(60);        // blank block was skipped
                            }
                            blocksdone[block] = 2; // block was written
                        }
                    } else {
                        Sleep(60);
                        blocksdone[block] = 1;  // unselected block was skipped
                    }
                hdc = GetDC(hwnd);
                ShowBarSegment(hdc, block);     // show block status in progress bar
                ReleaseDC(hwnd, hdc);
                }
            }
               ExitProgMode();
        } while (rc == IDRETRY);
           SetCursor(NULL);                        // show normal mouse pointer
    }
 	EnableWindow(GetDlgItem(hwnd, IDB_WCHP), TRUE);
	EnableWindow(GetDlgItem(hwnd, IDB_EB4W), TRUE);     // enable buttons after writing
	EnableWindow(GetDlgItem(hwnd, IDB_EXIT), TRUE);
}


/******************************************************************************
 *                     Map Bytes Window message handler                       *
 ******************************************************************************/

LRESULT CALLBACK MapWinProc (HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lparam)
{
    PAINTSTRUCT ps;
    HDC hdc;
    int line, addr, offset, x, y;
    wchar_t textbuff[80];

    switch (Msg) {
    case WM_CREATE:
        break;

    case WM_DESTROY:
        mapwindow = NULL;
        break;

    case WM_PAINT:
        hdc = BeginPaint(hwnd, &ps);
		if (hexfont) SelectObject(hdc, hexfont);
        for (line = 0; line < 16; line++) {
            if (line & 1) {
                SetBkColor(hdc, RGB(0xf0, 0xf0, 0xf0));
            }else {
                SetBkColor(hdc, RGB(0xff, 0xff, 0xff));
            }
            x = 0;
            y = line*16;
            addr = line*16;
			wsprintfW(textbuff, L"%04X:  ", (mapblock*256)+addr); // hex address
            SetTextColor(hdc, RGB(0x00, 0x00, 0x00)); // black
            TextOutW(hdc, x, y, textbuff, 6);
			for (offset = 0; offset < 16; offset++){
                if (bytemap[addr+offset] == 0xFF) {
                    SetTextColor(hdc, RGB(0x00, 0x80, 0x00)); // green
                    TextOutW(hdc, x+(5*8)+(offset*(3*8)) , y, L" FF  ", 5);  // byte is 0xFF (blank, erased)
                } else {
                    SetTextColor(hdc, RGB(0xE0, 0x00, 0x00)); // red
                    TextOutW(hdc, x+(5*8)+(offset*(3*8))-1, y, L" ??  ", 5); // byte is 0x00~0xFE (not blank)
                }
            }
        }
        EndPaint(hwnd, &ps);
        break;

    default:
        return DefWindowProcW(hwnd, Msg, wParam, lparam);
    }
    return 0;
}


//-----------------------------------------------------------------------------
//           Clicked on a block button in 'Blank Check' window
//-----------------------------------------------------------------------------
void DoMap (HWND hwnd, int block)
{
    RECT wr;
    int x, y, rc;
    wchar_t title[30];

    mapblock = block;
    wsprintfW(title, L"Byte Map - Block %d", mapblock);
    HCURSOR busy_pointer = LoadCursorW(NULL, IDC_WAIT);
    SetCursor(busy_pointer);
    do {
        rc = MapBytes(mapblock);           // get erasure state of each byte in block
        if (rc != 0) {
            rc = Request(hwnd, L"Target not responding", L"Byte Map - Error!",
                         20, 50, MB_ICONWARNING|MB_RETRYCANCEL);
        }
    } while (rc == IDRETRY);
    if (rc == IDCANCEL) {                 // MapBytes failed, so...
            if (mapwindow) {
                DestroyWindow(mapwindow); // destroy Byte Map window if it exists
            }
            return;                       // quit
    } else {
        SetCursor(NULL);
        // get calling window coordinates
        GetWindowRect(hwnd, &wr);
        x = max(0, wr.left+20);
        y = max(0, wr.top-25);
        if (mapwindow) {                  // Byte Map window already created?
            SetWindowText(mapwindow, title);
            InvalidateRect(mapwindow, NULL, FALSE); // yes, force display update
        } else {
            mapwindow = CreateWindowExW(0,          // no, create window
                                        L"MapWindow", title,
                                        WS_VISIBLE | WS_CAPTION | WS_SYSMENU,
                                        x, y, 432, 281,
                                        NULL, NULL, g_hinst, 0);
	}
    }
}


/******************************************************************************
 *                 Blank check Window Message Handler                         *
 ******************************************************************************/
LRESULT CALLBACK BlankWinProc (HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
    HWND blockbutton;
    HDC hdc;
    HBITMAP oldbmp;
    HBRUSH brush;
    HPEN pen;
    RECT rc;
    HFONT oldfont;
    int i;

	switch (Msg) {
	case WM_CREATE:
        // create check blocks groupbox and checkboxes
#define CBX 15
#define CBY 10
		CreateWindowExW(0, L"button", L"Block Erasure Status",
                        WS_CHILD | WS_VISIBLE | BS_GROUPBOX | BS_CENTER,
                        CBX, CBY, 440, 110, hwnd, (HMENU) IDB_BLKS, NULL, NULL);
        rc.left   = 0;
        rc.top    = 0;
        rc.right  = 13;
        rc.bottom = 13;

        hdc = GetDC(hwnd);

        memdc = CreateCompatibleDC(hdc);
        redbmp = CreateCompatibleBitmap(hdc, 13, 13);
        oldbmp = (HBITMAP) SelectObject(memdc, redbmp);
        brush = CreateSolidBrush(RGB(0xff, 0x00, 0x00));
        FillRect(memdc, &rc, brush);     // fill red
        DeleteObject(brush);
        pen = CreatePen(PS_SOLID, 2, RGB(0x00,0x00,0x00));
        SelectObject(memdc, pen);
	    MoveToEx(memdc, 2, 2, NULL);
        LineTo(memdc, 10, 10);           // draw a cross
	    MoveToEx(memdc, 2, 10, NULL);
        LineTo(memdc, 10, 2);
        DeleteObject(pen);

        grnbmp = CreateCompatibleBitmap(hdc, 13, 13);
        SelectObject(memdc, grnbmp);
        brush = CreateSolidBrush(RGB(0x00, 0xd0, 0x00));
        FillRect(memdc, &rc, brush);    // fill green
        DeleteObject(brush);
        pen = CreatePen(PS_SOLID, 2, RGB(0x00,0x00,0x00));
        SelectObject(memdc, pen);
	    MoveToEx(memdc, 2, 5, NULL);
        LineTo(memdc, 5, 10);           // draw a tick
        LineTo(memdc, 10, 2);
        DeleteObject(pen);

        SelectObject(memdc, oldbmp);

        for (i = 0; i < 16; i++) {
            if (i < (flashsize/256)) {
                blockbutton = CreateWindowExW(0, L"button", 0,
                                              WS_VISIBLE | WS_CHILD | BS_TOP | BS_BITMAP,
                                              CBX+18+((i&7)*52), CBY+30+((i/8)*40), 20, 20,
                                              hwnd, (HMENU)(IDB_BLK0+i) , NULL, NULL);
                if (blockstates[i] == 2) {
                    SendMessage(blockbutton, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)redbmp);
                }
                if (blockstates[i] == 1) {
                    SendMessage(blockbutton, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)grnbmp);
                }
            } else {
                blockbutton = CreateWindowExW(0, L"button", 0,
                                              WS_VISIBLE | WS_CHILD,
                                              CBX+18+((i&7)*52), CBY+30+((i/8)*40), 20, 20,
                                              hwnd, (HMENU)(IDB_BLK0+i) , NULL, NULL);
                SendMessage(blockbutton, BM_SETSTATE, 1, 0);
                EnableWindow(blockbutton, FALSE);
            }
        }
		// create 'exit' button
	    CreateWindowExW(0, L"button", L"Exit",
                        WS_VISIBLE | WS_CHILD,
                        385, 155, 70, 25,  // x, y, w, h
                        hwnd, (HMENU) IDB_EXIT, NULL, NULL);
		return 0;

	case WM_COMMAND:
		if (HIWORD(wParam) == BN_CLICKED) { // a button was pressed
            i = LOWORD(wParam);
			switch (i) {
	      	case IDB_EXIT: 			        // 'Exit' button
                EnableWindow(mainwindow,TRUE);
                DestroyWindow(hwnd);
                break;
 			default:
                if (i >= IDB_BLK0 && i <= IDB_BLK15) {
                    DoMap(hwnd, i - IDB_BLK0); // map the block
                }
                break;
			}
		}
		return 0;

	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps );
		SetBkMode( hdc, TRANSPARENT );
        for (i=0; i < 16; i++) {
            if (i < (flashsize/256)) {
            SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT));
            TextOutW(hdc, CBX+41+((i&7)*52), CBY+32+((i/8)*40), blocknames[i], lstrlenW(blocknames[i]));
            } else {
                SetTextColor(hdc, GetSysColor(COLOR_HIGHLIGHTTEXT));
                TextOutW(hdc, CBX+42+((i&7)*52), CBY+33+((i/8)*40), blocknames[i], lstrlenW(blocknames[i]));
                SetTextColor(hdc, GetSysColor(COLOR_GRAYTEXT));
                TextOutW(hdc, CBX+41+((i&7)*52), CBY+32+((i/8)*40), blocknames[i], lstrlenW(blocknames[i]));
            }

        }
        SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT));
        oldbmp = (HBITMAP) SelectObject(memdc, redbmp);
        BitBlt(hdc, 37, 129, 13, 13, memdc, 0, 0, SRCCOPY);
        SelectObject(memdc, oldbmp);
        TextOutW(hdc, 54, 128, L"= Not blank", 11);
        oldbmp = (HBITMAP) SelectObject(memdc, grnbmp);
        BitBlt(hdc, 192, 129, 13, 13, memdc, 0, 0, SRCCOPY);
        SelectObject(memdc, oldbmp);
        TextOutW(hdc, 209, 128, L"= Blank", 7);
		oldfont = SelectObject(hdc, sanfont);
        TextOutW(hdc, CBX, 165, L"Click on any block to read the status of each byte in it...", 59);
        SelectObject(hdc, oldfont);
		EndPaint(hwnd, &ps);
		return 0;

	case WM_DESTROY:
	    Request(mainwindow, L"Turn target power OFF", L"Blank Check",
                                                 180, 220, MB_ICON_SW_OFF|MB_OK);
        CloseHandle(hComm);
        DeleteDC(memdc);
        DeleteObject(grnbmp);
        DeleteObject(redbmp);
        blankwindow = NULL;
        return 0;

	default:
		return DefWindowProcW(hwnd, Msg, wParam, lParam);
	}
}



/******************************************************************************
 *                      Erase Window Message Handler                          *
 ******************************************************************************/
LRESULT CALLBACK EraseWinProc (HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
    HWND checkbox;
    int i;

	switch (Msg) {
	case WM_CREATE:
        // create erase blocks groupbox and checkboxes
#define EBX 20
#define EBY 10
		CreateWindowExW(0, L"button", L"blocks to erase",
                        WS_CHILD | WS_VISIBLE | BS_GROUPBOX | BS_CENTER,
                        EBX, EBY, 440, 90, hwnd, (HMENU) IDB_BLKS, NULL, NULL);

        for (i = 0; i < 16; i++) {
            checkbox = CreateWindowExW(0, L"button", blocknames[i],
                           WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX,
                           EBX+26+((i&7)*50), EBY+20+((i/8)*30), 40, 30,
                           hwnd, (HMENU)(IDB_BLK0+i) , NULL, NULL);
            if (i >=(flashsize/256)) EnableWindow(checkbox, FALSE);
        }
		// create button for 'Erase Chip' option
       	CreateWindowExW(0, L"button", L"Erase Chip",
                        WS_VISIBLE | WS_CHILD,
                        EBX, 125, 100, 25,	// x ,y ,w ,h
   		hwnd, (HMENU) IDB_ECHP, NULL, NULL);

	    // create 'Erase Blocks' button
       	CreateWindowExW(0,L"button", L"Erase Blocks",
                        WS_VISIBLE | WS_CHILD,
                        195, 125, 110, 25,	// x ,y ,w ,h
   		hwnd, (HMENU) IDB_EBLK, NULL, NULL);

		// create 'exit' button
	    CreateWindowExW(0, L"button", L"Exit",
                        WS_VISIBLE | WS_CHILD,
                        390, 125, 70, 25,  // x, y, w, h
                        hwnd, (HMENU) IDB_EXIT, NULL, NULL);
		return 0;

	case WM_COMMAND:
		if (HIWORD(wParam) == BN_CLICKED) { // a button was pressed
			switch (LOWORD(wParam)) {
	   		case IDB_ECHP:   			    // 'Erase Chip' button
                EraseChip(hwnd);
                break;
	       	case IDB_EBLK:                  // 'Erase Blocks' button
                EraseBlocks(hwnd);
                break;
	      	case IDB_EXIT: 			        // 'Exit' button
                EnableWindow(mainwindow,TRUE);
                DestroyWindow(hwnd);
                break;
			default:
                break;
			}
		}
		return 0;

	case WM_PAINT:
		BeginPaint(hwnd, &ps );
        // do drawing stuff here!
		EndPaint(hwnd, &ps );
		return 0;

	case WM_DESTROY:
        Request(hwnd, L"Turn target power OFF", L"Erase Flash",
                                          EBRX, EBRY, MB_ICON_SW_OFF|MB_OK);
        CloseHandle(hComm);
	    erasewindow = NULL;
        return 0;

	default:
		return DefWindowProcW(hwnd, Msg, wParam, lParam);
	}
}

/******************************************************************************
 *                      Write Window message handler                          *
 ******************************************************************************/

LRESULT CALLBACK WriteWinProc (HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    HDC hdc;
	PAINTSTRUCT ps;
    HWND checkbox;
    int i;

	switch (Msg) {
	case WM_CREATE:
        // create erase blocks groupbox and checkboxes
#define WBX 25
#define WBY 10
		CreateWindowExW(0, L"button", L"blocks to write",
                        WS_CHILD | WS_VISIBLE | BS_GROUPBOX | BS_CENTER,
                        WBX, WBY, 440, 90, hwnd, (HMENU) IDB_BLKS, NULL, NULL);
        for (i = 0; i < 16; i++) {
            checkbox = CreateWindowExW(0, L"button", blocknames[i],
                                       WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX,
                                       WBX+26+((i&7)*50), WBY+20+((i/8)*30), 40, 30,
                                       hwnd, (HMENU)(IDB_BLK0+i) , NULL, NULL);
            if (i >=(flashsize/256)) EnableWindow(checkbox, FALSE);
        }
		// create checkbox for 'Erase before Write' option
	    CreateWindowExW(0, L"button", L"erase before writing",
                        WS_VISIBLE | WS_CHILD | BS_AUTOCHECKBOX,
                        180, 165, 170, 25,	// x ,y ,w ,h
                        hwnd, (HMENU) IDB_EB4W, NULL, NULL);
        if (eraseflag) {
            CheckDlgButton(hwnd, IDB_EB4W, BST_CHECKED); // tick the box
        }

		// create 'write chip' button
	    CreateWindowExW(0,L"button", L"Write Chip",
                        WS_VISIBLE | WS_CHILD,
                        WBX, 210, 90, 25,    // x ,y ,w ,h
                        hwnd, (HMENU) IDB_WCHP, NULL, NULL);

		// create 'write blocks' button
	    CreateWindowExW(0,L"button", L"Write Blocks",
                        WS_VISIBLE | WS_CHILD,
                        200, 210, 110, 25,	// x ,y ,w ,h
                        hwnd, (HMENU) IDB_WBLK, NULL, NULL);

		// create 'exit' button
	    CreateWindowExW(0, L"button", L"Exit",
                        WS_VISIBLE | WS_CHILD ,
                        395, 210, 70, 25,  // x ,y ,w ,h
                        hwnd, (HMENU) IDB_EXIT, NULL, NULL);
		return 0;

	case WM_COMMAND:
		if (HIWORD(wParam) == BN_CLICKED) { // a button was pressed
			switch (LOWORD(wParam)) {
	   		case IDB_WCHP:  			    // 'write chip' button
                WriteChip(hwnd);
                break;
	   		case IDB_WBLK:  			    // 'write blocks' button
                WriteBlocks(hwnd);
                break;
	       	case IDB_EB4W:                  // 'erase before write' tickbox
                if (IsDlgButtonChecked(hwnd, IDB_EB4W) == BST_CHECKED) {
                    eraseflag = 1;
                } else {
                    eraseflag = 0;
                }
                break;
	       	case IDB_EXIT: 			        // 'Exit' button
                EnableWindow(mainwindow,TRUE);
			    DestroyWindow(hwnd);
                break;
			default:
                break;
			}
		}
		return 0;

	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);
        ShowProgBar(hdc);
		EndPaint(hwnd, &ps);
		return 0;

	case WM_DESTROY:
	    Request(hwnd,L"Turn target power OFF", L"Write Flash",
                                           SWX, SWY, MB_ICON_SW_OFF|MB_OK);
        CloseHandle(hComm);
        writewindow = NULL;
        return 0;

	default:
		return DefWindowProcW(hwnd, Msg, wParam, lParam);
	}
}

#define PBX WBX
#define PBY 127
#define PBW 438
#define PBH 20

void ShowBarSegment(HDC hdc, int block)
{
    RECT wr;
    HBRUSH brush = 0;
    int numblocks;

    numblocks = flashsize/256;
    wr.bottom = PBY+PBH;
    wr.left = PBX+1+(block*(PBW-1))/numblocks;
    wr.right = PBX+1+((block+1)*(PBW-1))/numblocks;
    wr.top = PBY+1;
    switch (blocksdone[block]) {
    case 2:
        brush = CreateSolidBrush(RGB(0x30, 0x60, 0xC0)); // color of written block
        break;
    case 1:
        brush = CreateSolidBrush(RGB(0xC0, 0xC0, 0xE0)); // color of skipped block
        break;
    case 3:
        brush = CreateSolidBrush(RGB(0xF0, 0x00, 0x00)); // color of failed block
        break;
    default:
        brush = (HBRUSH) (COLOR_BTNFACE+1);              // background color
    }
    FillRect(hdc, &wr, brush);
    if (brush) DeleteObject(brush);
}

void ShowProgBar (HDC hdc)
{
    HPEN pen;
    int i, x, y;

    pen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNSHADOW));
    SelectObject(hdc, pen);
	MoveToEx(hdc, PBX, PBY+PBH, NULL);
    LineTo(hdc, PBX, PBY);              // draw left & top border
    LineTo(hdc, PBX+PBW, PBY);
    DeleteObject(pen);

    pen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNHIGHLIGHT));
    SelectObject(hdc, pen);
    MoveToEx(hdc, PBX, PBY+PBH, NULL);
    LineTo(hdc, PBX+PBW, PBY+PBH);      // draw bottom & right border
    LineTo(hdc, PBX+PBW, PBY);
    DeleteObject(pen);

    pen = CreatePen(PS_SOLID, 1, RGB(0, 0, 0));
    SelectObject(hdc, pen);
    SetBkMode(hdc, TRANSPARENT);
    SetTextAlign(hdc, TA_CENTER);
    for (i = 0; i <= flashsize; i+=256) {
        x = PBX+((i*(PBW-1))/flashsize);
        y = PBY-1;
        MoveToEx(hdc, x, y, NULL);
        if ((i&1023) == 0) {      // draw kbyte markers on scale
           TextOutW(hdc, x, y-13, kbytenames[(i/1024)], lstrlenW(kbytenames[(i/1024)]));
        } else {
           LineTo(hdc, x, y-6);  // draw block markers on scale
        }
    }
    DeleteObject(pen);

    for (i = 0; i < (flashsize/256); i++) {
        ShowBarSegment(hdc, i); // show progress bar segments
    }
}


/******************************************************************************
 *            Hexadecimal Display Window Message Handler                      *
 ******************************************************************************/

#define PAGELINES 32	// lines per page (in display area)

long MaxFontHeight;

LRESULT CALLBACK HexWinProc (HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps; 		// painting structure
    SCROLLINFO si; 			// scrollbar structure
    HDC hdc;

    int yPos;               // vertical scrolling position
    int nMax;               // last line

    int x, y;               // horizontal and vertical pixel coordinates
    int line;               // current line
    int FirstLine;          // first line to be redrawn in hex display area
    int LastLine;           // last line to be redrawn in hex display area

	int addr, offset;       // address of byte in code buffer
	int byte;               // byte in code buffer
    wchar_t textbuff[80];   // text buffer to take printf formatted ouput

	switch (Msg) {
   	case WM_CREATE: 		// Window has just been created
   	    hdc = GetDC(hwnd);
		// switch to fixed width font
 		if (hexfont) SelectObject(hdc, hexfont);
		// Set vertical scrolling range and page size
    	si.cbSize = sizeof(si);
    	si.fMask  = SIF_RANGE | SIF_PAGE;
        si.nMin   = 0;
        si.nMax   = (flashsize/16)-1;
        si.nPage  = PAGELINES;
        SetScrollInfo(hwnd, SB_VERT, &si, TRUE);
        ReleaseDC(hwnd, hdc);
        return 0;

    case WM_VSCROLL: 	        	// Vertical Scroll Bar action
	    si.cbSize = sizeof (si);
        si.fMask  = SIF_ALL;		// Get all vertical scroll bar info
        GetScrollInfo(hwnd, SB_VERT, &si);
        yPos = si.nPos;				// Save old position for comparison
        switch (LOWORD(wParam)) {
      	case SB_TOP:			    // pressed HOME keyboard key
            si.nPos   = 0;
            break;
       	case SB_BOTTOM:				// pressed END keyboard key
            si.nPos = si.nMax;
            break;
       	case SB_LINEUP:				// clicked top arrow
            si.nPos -= 1;
            break;
       	case SB_LINEDOWN:			// clicked bottom arrow
            si.nPos += 1;
           	break;
       	case SB_PAGEUP:				// clicked above scroll box
           	si.nPos -= si.nPage;
           	break;
       	case SB_PAGEDOWN:			// clicked below scroll box
            si.nPos += si.nPage;
            break;
       	case SB_THUMBTRACK:			// dragged scroll box
            si.nPos = si.nTrackPos;
            break;
	    default:		    		// ignore other scroll bar actions
            break;
	    }

	// Set the position and then retrieve it.  Due to adjustments
	// by Windows it may not be the same as the value set.
    si.fMask  = SIF_POS | SIF_RANGE;
    si.nMin   = 0;
    si.nMax   = (flashsize/16)-1;
    SetScrollInfo(hwnd, SB_VERT, &si, TRUE);
    GetScrollInfo(hwnd, SB_VERT, &si);

	// If the position has changed, scroll the window and update it.
    if (si.nPos != yPos) {
        ScrollWindow(hwnd, 0, 16 * (yPos - si.nPos), NULL, NULL);
        UpdateWindow(hwnd);
    }
    return 0;

    case WM_PAINT:
        hdc = BeginPaint(hwnd, &ps);
		// Get vertical scroll bar position.
        si.cbSize = sizeof (si);
        si.fMask  = SIF_POS;
        GetScrollInfo(hexwindow, SB_VERT, &si);
        yPos = si.nPos;
        nMax = (flashsize/16)-1;
		// calculate which lines need to be displayed
        FirstLine = max(0, yPos + ps.rcPaint.top / 16);
        LastLine = min(nMax, yPos + ps.rcPaint.bottom / 16);

 	    if (hexfont) SelectObject(hdc, hexfont);
        for (line = FirstLine; line <= LastLine; line++) {
            x = 0;
            y = 16 * (line - yPos);

            if (line & 1) {
                SetBkColor(hdc, RGB(0xf0, 0xf0, 0xf0));
            }else {
                SetBkColor(hdc, RGB(0xff, 0xff, 0xff));
            }
			addr = line*16;
			// show hex address
			wsprintfW(textbuff, L"%04X:   ", addr);
            SetTextColor(hdc, RGB(0x00, 0x00, 0x00));
            TextOutW(hdc, x, y, textbuff, 8);
			// show 16 bytes of hex data
			for (offset = 0; offset < 16; offset++){
                byte = (int) CodeBuff[addr+offset];
				wsprintfW(textbuff, L"%02X  ", byte);
                if (byte == 0xFF) {
                    SetTextColor(hdc, RGB(0xA0, 0xA0, 0xA0));
                } else {
                    SetTextColor(hdc, RGB(0x00, 0x00, 0x00));
                }
                TextOutW(hdc, x+(8*6)+(offset*(8*3)), y, textbuff, 4);
            }
        }

		// Indicate that painting is finished.
        EndPaint(hwnd, &ps);
        return 0;

	default:
		return DefWindowProcW(hwnd, Msg, wParam, lParam);

	}
}

//-----------------------------------------------------------------------------
//        open, read and translate hex file into binary memory buffer
//-----------------------------------------------------------------------------
long LoadFile (wchar_t* file)
{
	FILE *fp;
	long ln = -1;
    CodeSize = 0;
	if ((fp = _wfopen(file, L"rt")) != NULL) {
		ln = input_hexfile(fp, CodeBuff, sizeof(CodeBuff), &CodeSize);
		fclose(fp);
	}
    return ln;
}


//-----------------------------------------------------------------------------
//                Open File Dialog and Load Hex File
//-----------------------------------------------------------------------------
void OpenFileDialog (HWND hwnd)
{
  OPENFILENAME ofn;
  wchar_t szFile[MAX_PATH];
  long rc;

  ZeroMemory(&ofn, sizeof(ofn));
  ofn.lStructSize = sizeof(ofn);
  ofn.lpstrFile = szFile;
  ofn.lpstrFile[0] = '\0';
  ofn.hwndOwner = hwnd;
  ofn.nMaxFile = sizeof(szFile);
  ofn.lpstrFilter = L"Intel Hex (*.hex)\0*.hex\0Motorola S (*.s19)\0*.s19\0All Files (*.*)\0*.*\0";
  ofn.nFilterIndex = 1;
  ofn.lpstrInitialDir = NULL;
  ofn.lpstrFileTitle = NULL;
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

  if (GetOpenFileNameW(&ofn)) {
        memset(CodeBuff, 0xFF, sizeof(CodeBuff));   // clear the code buffer
        rc = LoadFile(ofn.lpstrFile);         		// load/translate hex file
        if (rc == 0 && CodeSize > 0) {
            lstrcpyW(filename, ofn.lpstrFile);      // get local copy of file name
            Truncate(filename, 80);                 // trim name for display
        } else {
            lstrcpyW(filename, L"no data!");        // bad Hex file!
        }
        SetWindowTextW(filenamebox, filename);      // show file name
	    InvalidateRect(mainwindow, NULL, FALSE);    // force display update
  }

  InvalidateRect(hexwindow, NULL, FALSE);           // force hex display update
}

#define TRX 210
#define TRY 120
//-----------------------------------------------------------------------------
//                          "Test" Button pressed
//-----------------------------------------------------------------------------
int TestComms (HWND hwnd)
{
    int rc; // requester return code
	rc = Request(hwnd, L"Ensure target power is OFF, then click OK to open COM port",
                     L"Test - open COM port", TRX, TRY, 0 | MB_OKCANCEL);
    if (rc == IDOK) {
        if (OpenComPort(hwnd)) {
            rc = Request(hwnd, L"Turn target power ON, then click OK to probe target",
                         L"Test - Probe Target", TRX, TRY, 0 | MB_ICON_SW_ON|MB_OKCANCEL);
            if (rc == IDOK) {
                do {
                    if (EnterProgMode()) {
                        rc = Request(hwnd, L"Probe Successful - Target found",
                                     L"Test - Passed", TRX, TRY, 0 | MB_OK);
                    } else {
                        rc = Request(hwnd, L"Target not responding", L"Test - Error!",
                                     TRX, TRY, MB_ICONWARNING | MB_RETRYCANCEL);
                    }
                    ExitProgMode();
                } while (rc == IDRETRY);
            Request(hwnd, L"Test complete.\nTurn target power OFF",
                    L"Test - Close Com Port", TRX, TRY, MB_ICON_SW_OFF|MB_OK);
            }
        CloseHandle(hComm);			// release com port
        }
    }
	return 0;
}

//-----------------------------------------------------------------------------
//                          "Check" Button pressed
//-----------------------------------------------------------------------------
int DoBlank (HWND hwnd)
{
    RECT wr;
    int block, x, y, rc;

    if (OpenComPort(hwnd)) {
        rc = Request(hwnd, L"Turn target power ON", L"Blank Check",
                                                       180, 220, MB_ICON_SW_ON|MB_OKCANCEL);
        if (rc == IDOK) {
            do {
                rc = 0;
                if (EnterProgMode()) {
                    for (block = 0; block < 16; block++) {
                        if (block < (flashsize/256)) {
                            blockstates[block] = blankcheck(block);
                        } else {
                            blockstates[block] = 0; // not a valid block
                        }
                    }
                } else {
                    rc = Request(hwnd, L"Target not responding", L"Blank Check - Error!",
                                              180, 220, MB_ICONWARNING | MB_RETRYCANCEL);
                }
                ExitProgMode();
            } while (rc == IDRETRY);

            if (rc != IDCANCEL) {
                // disable main window while blank checking
                EnableWindow(hwnd,FALSE);
                // get main window coordinates
                GetWindowRect(hwnd, &wr);
                x = wr.left;
                y = wr.top;
                // create blank check window
                blankwindow = CreateWindowExW(0, L"BlankWindow",
                                            L"78K0s/Kx1 - Blank Check",
                                            WS_VISIBLE | WS_CAPTION,
                                            x+60, y+200, 480, 220,
                                            NULL, NULL, g_hinst, NULL);
                if (mapwindow) {
                    DestroyWindow(mapwindow); // destroy Byte Map window if it exists
                }
                return 0;
            }
        }
        CloseHandle(hComm);
    }
	return 0;
}


#define ERX 180
#define ERY 250

//-----------------------------------------------------------------------------
//                          "Erase" Button pressed
//-----------------------------------------------------------------------------
int DoErase (HWND hwnd)
{
    RECT wr;
    int rc, x, y;

	// disable main window while Erase window is active
	EnableWindow(hwnd,FALSE);
	// get main window coordinates
    GetWindowRect(hwnd, &wr);
	x = wr.left;
	y = wr.top;

    if (OpenComPort(hwnd)) {
        rc = Request(hwnd, L"Turn Target power ON", L"Erase Flash",
	                                                    170, 260, MB_ICON_SW_ON|MB_OKCANCEL);
        if (rc != IDCANCEL) {
            // create erase window
            erasewindow = CreateWindowExW(
            0,								// extended style
            L"EraseWindow",					// class name
            L"78K0s/Kx1 - Erase Flash",     // window name
            WS_VISIBLE | WS_CAPTION,        // style
            x+60, y+210, 487, 190,			// x, y, w, h
            NULL, NULL, g_hinst, NULL);
            return 0;
        } else {
            CloseHandle(hComm);
        }
    }
    return 0;
}

//-----------------------------------------------------------------------------
//				         	"Write" Button pressed
//-----------------------------------------------------------------------------
int DoWrite (HWND hwnd)
{
    RECT wr;
    int i, x, y, rc;

	// disable main window while Write window is active
	EnableWindow(hwnd,FALSE);
	// get main window coordinates
    GetWindowRect(hwnd, &wr);
	x = wr.left;
	y = wr.top;

    if (OpenComPort(hwnd)) {
        rc = Request(hwnd, L"Turn Target power ON", L"Write Flash",
	                                                    170, 275, MB_ICON_SW_ON|MB_OKCANCEL);
        if (rc != IDCANCEL) {
            for (i=0;i<16;i++) {
                blocksdone[i] = 0;
            }
            // create programming window
            writewindow = CreateWindowExW(
            0,											// extended style
            L"WriteWindow",								// class name
            L"78K0s/Kx1 - Write Code to Flash",			// window name
            WS_VISIBLE | WS_CAPTION,	// style
            x+52, y+175, 500, 280,						// x, y, w, h
            NULL, NULL, g_hinst, NULL);

        } else {
            CloseHandle(hComm);
        }
    }
	return 0;
}

//------------------------------------------------------------------------------
//			             Save Settings to Registy
//------------------------------------------------------------------------------
int SaveSettings ()
{
// local variables used by registry functions
    wchar_t keybuff[40];                // buffer to hold subkey value (Unicode string)
    DWORD keylen = sizeof(keybuff);     // length of the buffer in bytes
    HKEY regkey;                        // handle to our key
    int rc;                             // return code from registry functions
    RECT wr;

    GetWindowRect(mainwindow, &wr);
	window_x = wr.left;
	window_y = wr.top;

// create our registry key (or just open it, if it already exists)
    rc = RegCreateKeyExW(HKEY_CURRENT_USER, L"Software\\78K0Sprog",
                         0, 0, 0, KEY_ALL_ACCESS, 0, &regkey, 0);
    if (rc == ERROR_SUCCESS) {

// create settings string and write it to our subkey
        wsprintfW((LPWSTR)keybuff, L"%d,%d,%d,%d,%s", flashsize, eraseflag, window_x, window_y, comportname);
        rc = RegSetValueExW(regkey, L"Settings", 0, REG_SZ, (BYTE *) &keybuff, keylen);
        RegCloseKey(regkey);
    }

	return rc;
}

//------------------------------------------------------------------------------
//			            Load Settings from Registy
//------------------------------------------------------------------------------
int LoadSettings ()
{
// local variables used by registry functions
    wchar_t keybuff[40];                // buffer to hold subkey value (Unicode string)
    DWORD keylen = sizeof(keybuff);     // length of the buffer in bytes
    DWORD keytype = 0;                  // actual key type (should be REG_SZ)
    HKEY regkey;                        // handle to our key
    int rc;                             // return code from registry functions

// open our registry key and extract settings from subkey string
     rc = RegOpenKeyExW(HKEY_CURRENT_USER, L"Software\\78K0Sprog", 0, KEY_QUERY_VALUE, &regkey);
     if (rc == ERROR_SUCCESS) {
        rc = RegQueryValueExW(regkey, L"Settings", 0, &keytype, (BYTE *) &keybuff, &keylen);
        RegCloseKey(regkey);
        if ((rc == ERROR_SUCCESS) && (keytype == REG_SZ)) {
            swscanf((LPWSTR)keybuff,L"%d,%d,%d,%d,%s", &flashsize, &eraseflag, &window_x, &window_y, comportname);
        }
    }
	return rc;
}


//----------------------------------------------------------------------------
//                              Requester
//----------------------------------------------------------------------------
/* Like a MessageBox, but with x y coordinates set relative to our window

   Waits for response before closing the requester and returning user's choice
*/
struct REQDATA {
    HWND     win;   // calling window
    LPCWSTR text;   // text to show in requester body
    int     type;   // type of requester (eg. MB_OKCANCEL)
};

int Request (HWND hwnd, LPCWSTR text, LPCWSTR title, INT x, INT y, INT type)
{
    RECT wr;

    if (hwnd) {
        EnableWindow(hwnd,FALSE); // disable calling window
        GetWindowRect(hwnd, &wr); // get calling window's coordinates
    }
    struct REQDATA req = {0};     // parameters to pass to our requester
    req.win  = hwnd;
    req.text = text;
    req.type = type;

    switch (type&0xF0) {
    case MB_ICONERROR:
        MessageBeep(MB_ICONERROR);
        break;
    case MB_ICONWARNING:
        MessageBeep(MB_ICONWARNING);
        break;
    case MB_ICONINFORMATION:
        MessageBeep(MB_ICONINFORMATION);
        break;
    case MB_ICONQUESTION:
        MessageBeep(MB_ICONQUESTION);
        break;
    default:
        MessageBeep(MB_OK);
    }
    if (hwnd) {
   	    x += wr.left;
	    y += wr.top;
    }
	reqwindow = CreateWindowExW(
        WS_EX_DLGMODALFRAME | WS_EX_TOPMOST,
        L"ReqWindow",		// class
        title,
        WS_VISIBLE | WS_CAPTION,
        x, y, 256, 140,	    // x,y,w,h
        NULL,      			// parent window
        NULL,           	// window menu id
        g_hinst,			// our instance
        &req         		// lParam
	);
	if (reqwindow == NULL) return 0;
 MSG Msg;
	while (GetMessageW(&Msg, NULL, 0, 0) > 0) {
        if (Msg.message == WM_USER) break;
		DispatchMessageW(&Msg);
	}
    return Msg.wParam;      // return choice (eg. IDOK)
}

/******************************************************************************
 *                     Requester Window message handler                       *
 ******************************************************************************/

LRESULT CALLBACK ReqWinProc (HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lparam)
{
    static int type;    // requester type
    static HWND win;    // calling window

    switch (Msg) {
    case WM_CREATE: {
        struct REQDATA* req = (struct REQDATA*)((LPCREATESTRUCT)lparam)->lpCreateParams;
        type = req->type;
        win = req->win;
        int buttons = type&0x0F;
        int txtx = 10;
        if (type > 16) {
            txtx = 50;   // make room for icon
        }
        CreateWindowExW(0, L"static", req->text,    // body text
                        WS_VISIBLE | SS_CENTER | WS_CHILD ,
                        txtx, 20, 240-txtx, 50,
                        hwnd, (HMENU) IDQ_OK, NULL, NULL);
        if ((buttons) == MB_OKCANCEL) {
            CreateWindowExW(0, L"button", L"OK",
                            WS_VISIBLE | WS_CHILD ,
                            20, 80, 60, 25,
                            hwnd, (HMENU) IDQ_OK, NULL, NULL);
            CreateWindowExW(0, L"button", L"Cancel",
                            WS_VISIBLE | WS_CHILD ,
                            170, 80, 60, 25,
                            hwnd, (HMENU) IDQ_CANCEL, NULL, NULL);
        } else if (buttons == MB_RETRYCANCEL) {
            CreateWindowExW(0, L"button", L"Retry",
                            WS_VISIBLE | WS_CHILD ,
                            20, 80, 60, 25,
                            hwnd, (HMENU) IDQ_RETRY, NULL, NULL);
            CreateWindowExW(0, L"button", L"Cancel",
                            WS_VISIBLE | WS_CHILD ,
                            170, 80, 60, 25,
                            hwnd, (HMENU) IDQ_CANCEL, NULL, NULL);
        } else {
            CreateWindowExW(0, L"button", L"OK",
                            WS_VISIBLE | WS_CHILD ,
                            100, 80, 60, 25,
                            hwnd, (HMENU) IDQ_OK, NULL, NULL);
        }
        break;
        }

    case WM_DESTROY:
        reqwindow = NULL;
        break;

    case WM_PAINT: {
        PAINTSTRUCT ps;
        HDC hdc = BeginPaint(hwnd, &ps);
        HICON my_icon = NULL;
        switch (type&0xF0) {
        case MB_ICONERROR:
            my_icon = LoadIconW(NULL, IDI_ERROR);
            break;
        case MB_ICONWARNING:
            my_icon = LoadIconW(NULL, IDI_WARNING);
            break;
        case MB_ICONINFORMATION:
            my_icon = LoadIconW(NULL, IDI_INFORMATION);
            break;
        case MB_ICONQUESTION:
            my_icon = LoadIconW(NULL, IDI_QUESTION);
            break;
        case MB_ICON_SW_ON:
            my_icon = LoadIconW(g_hinst, L"SW_ON");
            break;
        case MB_ICON_SW_OFF:
            my_icon = LoadIconW(g_hinst, L"SW_OFF");
            break;
        }
        if (my_icon) {
           DrawIcon(hdc, 10, 20, my_icon);
        }
        EndPaint(hwnd, &ps);
        break;
        }
    case WM_COMMAND:
        if (HIWORD(wParam) == BN_CLICKED) { // button pressed
            switch (LOWORD(wParam)) {
            case IDQ_OK:
                PostMessageW(mainwindow, WM_USER, IDOK, 0);
                break;
            case IDQ_RETRY:
                PostMessageW(mainwindow, WM_USER, IDRETRY, 0);
                break;
            case IDQ_CANCEL:
                PostMessageW(mainwindow, WM_USER, IDCANCEL, 0);
            }
        if (win) EnableWindow(win,TRUE);
        DestroyWindow(hwnd);
        return 0;
        }
        break;

	// all other messages get default Window procedures
    default:
        return DefWindowProcW(hwnd, Msg, wParam, lparam);
    }
    return 0;
}

/******************************************************************************
 *                      Main Window Message Handler                           *
 ******************************************************************************/

LRESULT CALLBACK MainWinProc(HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	switch (Msg) {
    case WM_CREATE: {
//------------------------------ file menu -----------------------------------
		HMENU hMenubar;
		HMENU hMenu;
		hMenubar = CreateMenu();
		hMenu = CreateMenu();
		AppendMenu (hMenubar, MF_POPUP, (UINT_PTR)hMenu, TEXT("&File"));
		AppendMenu (hMenu, MF_STRING, IDM_FILE_OPEN, TEXT("&Open"));
		SetMenu (hwnd, hMenubar);

//--------------------------- filename display -------------------------------
		filenamebox = CreateWindowExW(0, L"edit", L"no file", WS_CHILD |
                                      WS_VISIBLE | ES_READONLY,
                                      10, 3, 598, 14, hwnd, NULL,NULL,NULL);
        if (sanfont) SendMessage(filenamebox, WM_SETFONT, (WPARAM)sanfont, MAKELPARAM(1,0));

//--------------------------- Interface group --------------------------------
#define GRPX 485
#define INTY 20
		CreateWindowExW(0, L"button", L"Interface",
               WS_CHILD | WS_VISIBLE | BS_GROUPBOX |BS_CENTER,
               GRPX, INTY, 106, 110, hwnd, NULL, NULL, NULL);

		combox = CreateWindowExW(0, L"combobox", L"COM Port",
                   WS_CHILD | WS_VISIBLE |CBS_DROPDOWNLIST |CBS_HASSTRINGS,
                   GRPX+17, INTY+30, 70, 120, hwnd, NULL, NULL, NULL);

   		CreateWindowExW(0, L"button", L"Test",
                WS_VISIBLE | WS_CHILD ,
                GRPX+17, INTY+70, 70, 25,
                hwnd, (HMENU) IDB_TEST, NULL, NULL);

//-------------------- Initialize Com Port Dropdown Box --------------------------
		wchar_t szDevices[32767]; // buffer to hold dos device name list
		unsigned long dwchar_ts;  // number of chars in dos device name list
		wchar_t *comports[12];	  // array of pointers to com port names in buffer
		wchar_t *ptr = szDevices; // pointer to dos device name in buffer
		int	numports = 0;		  // total number of com ports found
		int port;				  // current com port number
        int selected = 0;         // selected dropdown box item number
		int i;

		// fill our buffer with dos device name strings
		dwchar_ts = QueryDosDeviceW(NULL, szDevices, 32765);
		// find any com port names in the buffer and add them to our com port list
		while (dwchar_ts && (numports < 12)) {
			if (swscanf(ptr, L"COM%d", &port) == 1) {
		    	if ((port > 0) && (port < 100)) {     // only accept ports 1~99
                    if (!strcmpw(ptr, comportname)) { // if port = our port then
                        selected = numports;          // it will be the selected item
                    }
					comports[numports] = ptr;	// insert name ptr into array
					numports++;			    	// another com port found!
				}
			}
			wchar_t *temp_ptr = wcschr(ptr, 0); // point to end of current string
			dwchar_ts -= (DWORD)((temp_ptr-ptr)/sizeof(wchar_t)+1);
			ptr = temp_ptr + 1; 			    // skip to next string in buffer
		}
        // add com port names to dropdown box list
       	for (i = 0; i < numports; i++ ) {
           	SendMessageW(combox, CB_ADDSTRING, 0, (LPARAM) comports[i]);
       	}
		// set selected dropdown box item
		SendMessageW(combox, CB_SETCURSEL, selected, 0) ;
		// get selected com port name
		SendMessageW(combox,CB_GETLBTEXT, selected,(LPARAM)comportname);

//----------------------- Mem Size groupbox and buttons ----------------------
#define MEMY 162
		CreateWindowExW(0, L"button", L"ROM Size",
                WS_CHILD | WS_VISIBLE | BS_GROUPBOX | BS_CENTER,
                GRPX, MEMY, 106, 106, hwnd, (HMENU) IDB_MEM, NULL, NULL);

		CreateWindowExW(0, L"button", L"1k",
                WS_CHILD | WS_VISIBLE | BS_AUTORADIOBUTTON,
                GRPX+30, MEMY+20, 60, 30, hwnd, (HMENU)IDB_1K , NULL, NULL);

		CreateWindowExW(0, L"button", L"2k",
                WS_CHILD | WS_VISIBLE | BS_AUTORADIOBUTTON,
                GRPX+30, MEMY+45, 60, 30, hwnd, (HMENU)IDB_2K , NULL, NULL);

		CreateWindowExW(0, L"button", L"4k",
                WS_CHILD | WS_VISIBLE | BS_AUTORADIOBUTTON,
                GRPX+30, MEMY+70, 60, 30, hwnd, (HMENU)IDB_4K , NULL, NULL);
// select initial memory size
        switch (flashsize) {
        case 1024:
            i = IDB_1K;
            break;
        case 2048:
            i = IDB_2K;
            break;
        case 4096:
            i = IDB_4K;
            break;
        }
       	CheckRadioButton(hwnd, IDB_1K, IDB_4K, i);

//------------------------- Flash group and buttons --------------------------
#define FLASHY 300
		CreateWindowExW(0, L"button", L"Flash",
                WS_CHILD | WS_VISIBLE | BS_GROUPBOX |BS_CENTER,
                GRPX, FLASHY, 106, 154, hwnd, NULL, NULL, NULL);

   		CreateWindowExW(0, L"button", L"Check",
                WS_VISIBLE | WS_CHILD ,
                GRPX+17, FLASHY+29, 70, 25,
                hwnd, (HMENU) IDB_BLANK, NULL, NULL);

   		CreateWindowExW(0, L"button", L"Erase",
                WS_VISIBLE | WS_CHILD ,
                GRPX+17, FLASHY+71, 70, 25,
                hwnd, (HMENU) IDB_ERASE, NULL, NULL);

		CreateWindowExW(0, L"button", L"Write",
                WS_VISIBLE | WS_CHILD ,
                GRPX+17, FLASHY+113, 70, 25,
                hwnd, (HMENU) IDB_WRITE, NULL, NULL);

 		CreateWindowExW(0, L"button", L"Quit",
                WS_VISIBLE | WS_CHILD ,
                GRPX+17, FLASHY+190, 70, 25,
                hwnd, (HMENU) IDB_QUIT, NULL, NULL);
    	break;
	}
	case WM_PAINT: {
			PAINTSTRUCT ps;
			BeginPaint(hwnd, &ps);
            // do drawing stuff here!
			EndPaint(hwnd, &ps);
		break;
	}
   	case WM_COMMAND: {
	    if (wParam == IDM_FILE_OPEN) {	// open file
   	          OpenFileDialog(hwnd);
		}
		if (HIWORD(wParam) == BN_CLICKED) {
			switch (LOWORD(wParam)) {
            case IDB_1K:                // 1K memory size selection
                if (flashsize!=1024) {
                    flashsize=1024;
                    PostMessageW(hexwindow, WM_VSCROLL, SB_TOP, 0);
                }
                break;
            case IDB_2K:                // 2K memory size selection
                if (flashsize!=2048) {
                    flashsize=2048;
                    PostMessageW(hexwindow, WM_VSCROLL, SB_TOP, 0);
                }
                break;
            case IDB_4K:                // 4K memory size selection
                if (flashsize!=4096) {
                    flashsize=4096;
                    PostMessageW(hexwindow, WM_VSCROLL, SB_TOP, 0);
                }
                break;
            case IDB_TEST:
                TestComms(hwnd);	    // test comms
                break;
            case IDB_BLANK:
                DoBlank(hwnd);	        // blank check
                break;
            case IDB_ERASE:
                DoErase(hwnd);		    // erase flash
                break;
            case IDB_WRITE:
                DoWrite(hwnd);		    // write flash
                break;
            case IDB_QUIT:
                SaveSettings();
                DestroyWindow(hwnd);    // quit
                break;
            default:
                break;
			}
		}
		if (HIWORD(wParam) == CBN_SELCHANGE) {  // com port selection
			int selected = SendMessageW(combox, CB_GETCURSEL, 0, 0);
			// get name of selected com port
			SendMessageW(combox,CB_GETLBTEXT, selected,(LPARAM)comportname);
			// remove focus from dropdown box
			SetFocus(hwnd);
        }
        break;
	}
    case WM_CLOSE:                      // close window
	    SaveSettings();
        DestroyWindow(hwnd);
        break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	default:
		return DefWindowProcW(hwnd, Msg, wParam, lParam);
	}
	return 0;
}

//=============================================================================
//                    Windows Main Program Entry Point
//=============================================================================

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{

	g_hinst = hInstance;	// global copy of hInstance
	MSG Msg;				// our message
	WNDCLASSEX wc = {0};    // window class structure

    // nice looking san serif text font
    sanfont = CreateFontW(-12, 0, 0, 0, FW_NORMAL, 0, 0, 0, DEFAULT_CHARSET,
                             OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                             DEFAULT_QUALITY, FF_DONTCARE, L"Arial");

    // fixed width font for hex display
    hexfont = CreateFontW(16, 0, 0, 0, FW_NORMAL, 0, 0, 0, DEFAULT_CHARSET,
                             OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                             DEFAULT_QUALITY, FF_DONTCARE, L"Courier New");


    // Main window class
	wc.cbSize		  = sizeof(WNDCLASSEX);
	wc.lpfnWndProc	  = MainWinProc;
	wc.hInstance	  = hInstance;
	wc.hCursor		  = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground  = (HBRUSH)(COLOR_BTNFACE+1);
	wc.lpszMenuName   = NULL;
	wc.lpszClassName  = L"MainWindow";
	wc.hIcon		  = LoadIcon(g_hinst,L"APPICON");
	wc.hIconSm		  = LoadIcon(g_hinst,L"WINICON");
	if(!RegisterClassExW(&wc)) return 0;

    // Hex Display window class
	wc.lpfnWndProc	  = HexWinProc;
	wc.lpszClassName  = L"HexWindow";
	if(!RegisterClassExW(&wc)) return 0;

    // Write Flash window class
	wc.lpfnWndProc	  = WriteWinProc;
	wc.lpszClassName  = L"WriteWindow";
	if(!RegisterClassExW(&wc)) return 0;

    // Erase Flash window class
	wc.lpfnWndProc	  = EraseWinProc;
	wc.lpszClassName  = L"EraseWindow";
	if(!RegisterClassExW(&wc)) return 0;

    // Requester window class
	wc.lpfnWndProc	  = ReqWinProc;
	wc.lpszClassName  = L"ReqWindow";
	if(!RegisterClassExW(&wc)) return 0;

    // 'BlankCheck' window class
	wc.lpfnWndProc	  = BlankWinProc;
	wc.lpszClassName  = L"BlankWindow";
	if(!RegisterClassExW(&wc)) return 0;

    // 'Block Map' window class
	wc.lpfnWndProc	  = MapWinProc;
	wc.lpszClassName  = L"MapWindow";
	if(!RegisterClassExW(&wc)) return 0;

    LoadSettings();     // get settings from registry

	// create main window
	mainwindow = CreateWindowExW (
        WS_EX_CLIENTEDGE,
        L"MainWindow",
        L"78K0S/Kx1+ Programmer",
        WS_VISIBLE|WS_BORDER|WS_MINIMIZEBOX|WS_SYSMENU|WS_CAPTION,
        window_x,           // x
        window_y,           // y
        626,            	// width
        595,            	// height
        NULL,				// parent window
        NULL,				// window menu
        hInstance,			// our instance
        NULL				// lParam
	);

	if (mainwindow == NULL) return 0;

	// create hex display window
	hexwindow = CreateWindowExW(
        WS_EX_RIGHTSCROLLBAR,
        L"HexWindow",		// class
        L"Hex Display Window",
        WS_VISIBLE | WS_CHILD | WS_VSCROLL | WS_DLGFRAME,
        10,                 // x
        20,                 // y
        450,                // width
        (16*PAGELINES)+6,	// height
        mainwindow,			// parent window
        (HMENU)IDW_HEXD,	// window id
        hInstance,			// our instance
        NULL				// lParam
	);

	if (hexwindow == NULL) return 0;

// clear the code buffer to all 1's
    memset(CodeBuff, 0xFF, sizeof(CodeBuff));

// main loop: wait for messages and send them to our window procs
	while (GetMessageW(&Msg, NULL, 0, 0) > 0) {
		DispatchMessageW(&Msg);
	}
	return Msg.wParam;      // quit message received
}

